# -*- coding: utf-8 -*-

# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..utils.msgpack_for_db import MsgPackForDB


class DepositMeta(object):
    """
    DepositMeta Class implementing functions to serialize and deserialize.
    """
    _VERSION = 0

    def __init__(self, head_id: bytes = None, tail_id: bytes = None,
                 available_head_id_of_virtual_step: bytes = None, available_head_id_of_deposit: bytes = None,
                 expires_of_virtual_step: int = -1, expires_of_deposit: int = -1):
        self.version = self._VERSION
        self.head_id = head_id
        self.tail_id = tail_id
        self.available_head_id_of_virtual_step = available_head_id_of_virtual_step
        self.available_head_id_of_deposit = available_head_id_of_deposit
        self.expires_of_virtual_step = expires_of_virtual_step
        self.expires_of_deposit = expires_of_deposit

    @staticmethod
    def from_bytes(buf: bytes):
        """Converts DepositMeta in bytes into DepositMeta Object.

        :param buf: DepositMeta in bytes
        :return: DepositMeta Object
        """
        data: list = MsgPackForDB.loads(buf)

        deposit_meta = DepositMeta()
        deposit_meta.version = data[0]
        deposit_meta.head_id = data[1]
        deposit_meta.tail_id = data[2]
        deposit_meta.available_head_id_of_virtual_step = data[3]
        deposit_meta.available_head_id_of_deposit = data[4]
        deposit_meta.expires_of_virtual_step = data[5]
        deposit_meta.expires_of_deposit = data[6]

        return deposit_meta

    def to_bytes(self) -> bytes:
        """Converts DepositMeta object into bytes.

        :return: DepositMeta in bytes
        """
        data: list = [self.version, self.head_id, self.tail_id,
                      self.available_head_id_of_virtual_step, self.available_head_id_of_deposit,
                      self.expires_of_virtual_step, self.expires_of_deposit]
        return MsgPackForDB.dumps(data)

    def __eq__(self, other) -> bool:
        """operator == overriding

        :param other: (DepositMeta)
        """
        return isinstance(other, DepositMeta) \
            and self.version == other.version \
            and self.head_id == other.head_id \
            and self.tail_id == other.tail_id \
            and self.available_head_id_of_virtual_step == other.available_head_id_of_virtual_step \
            and self.available_head_id_of_deposit == other.available_head_id_of_deposit \
            and self.expires_of_virtual_step == other.expires_of_virtual_step \
            and self.expires_of_deposit == other.expires_of_deposit

    def __ne__(self, other) -> bool:
        """operator != overriding

        :param other: (DepositMeta)
        """
        return not self.__eq__(other)



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict
from collections.abc import MutableMapping
from copy import copy
from typing import Optional, List

from ..base.block import Block
from ..base.exception import DatabaseException, AccessDeniedException
from ..icx import IcxStorage
from ..utils import sha3_256, to_camel_case


class BatchValue:
    def __init__(self, value: Optional[bytes], include_state_root_hash: bool):
        self._value: bytes = value
        self._include_state_root_hash: bool = include_state_root_hash

    @property
    def value(self):
        return self._value

    @property
    def include_state_root_hash(self):
        return self._include_state_root_hash

    def to_dict(self, casing: Optional[callable] = None) -> dict:
        new_dict = {}
        for key, value in self.__dict__.items():
            if key.startswith("_"):
                key = key[1:]
            new_dict[casing(key) if casing else key] = value

        return new_dict


class TransactionBatchValue(BatchValue):
    def __init__(self, value: Optional[bytes], include_state_root_hash: bool, tx_index: int = -1):
        super().__init__(value, include_state_root_hash)
        self._tx_index: int = tx_index

    @property
    def tx_index(self) -> int:
        """
        Transaction index information for debugging
        index -1 means deploying score or the data being been recorded outside of the transaction
        :return:
        """
        # Fixme: Correct tx index should be set on deploying score (not -1)
        return self._tx_index

    def __repr__(self):
        return f'TransactionBatchValue({self.value.hex()}, {self.include_state_root_hash}, {self.tx_index})'

    def __eq__(self, other: 'TransactionBatchValue'):
        return self.value == other.value and \
               self.include_state_root_hash == other.include_state_root_hash and \
               self.tx_index == other.tx_index


class BlockBatchValue(BatchValue):
    def __init__(self, value: Optional[bytes], include_state_root_hash: bool, tx_indexes: List[int]):
        super().__init__(value, include_state_root_hash)
        self._tx_indexes: List[int] = tx_indexes

    @property
    def tx_indexes(self) -> List[int]:
        return copy(self._tx_indexes)

    def __repr__(self):
        return f'BlockBatchValue({self.value.hex()}, {self.include_state_root_hash}, {self.tx_indexes})'

    def __eq__(self, other: 'BlockBatchValue'):
        return self.value == other.value and \
               self.include_state_root_hash == other.include_state_root_hash and \
               self.tx_indexes == other.tx_indexes


def digest(ordered_dict: OrderedDict):
    # items in data MUST be byte-like objects
    data = []

    for key, tx_batch_value in ordered_dict.items():
        if tx_batch_value.include_state_root_hash is True:
            value: bytes = tx_batch_value.value
        else:
            continue
        data.append(key)
        if value is not None:
            data.append(value)
    value: bytes = b'|'.join(data)
    return sha3_256(value)


class Batch(OrderedDict):
    def __init__(self):
        super().__init__()

    def digest(self) -> bytes:
        """Create sha3_256 hash value with included updated states

        How to create a hash value:
        hash_value = sha3_256(b'key0|value0|key1|value1|...)

        case1: value1 is None,
            hash_value = sha3_256(b'key0|value0|key1|key2|value2|...)

        case2: value1 = b''
            hash_value = sha3_256(b'key0|value0|key1||value2|...)

        :return: sha3_256 hash value
        """
        # items in data MUST be byte-like objects
        return digest(self)


class TransactionBatch(MutableMapping):
    """Contains the states changed by a transaction.

    key: Score Address
    value: IconScoreBatch
    """

    def __init__(self, tx_hash: Optional[bytes] = None) -> None:
        """Constructor

        :param tx_hash: tx_hash
        """
        super().__init__()
        self.hash = tx_hash
        self._call_batches = [OrderedDict()]

    def __getitem__(self, item):
        for call_batch in reversed(self._call_batches):
            if item in call_batch:
                return call_batch[item]

        return None

    def __setitem__(self, key, value):
        assert isinstance(value, TransactionBatchValue)

        call_batch: OrderedDict = self._call_batches[-1]
        call_batch[key] = value

    def __delitem__(self, key):
        raise DatabaseException('delete item is not allowed')

    def __contains__(self, item):
        for call_batch in self._call_batches:
            if item in call_batch:
                return True

        return False

    def __iter__(self):
        for call_batch in self._call_batches:
            for key in call_batch:
                yield key

    def __len__(self):
        length = 0

        for call_batch in self._call_batches:
            length += len(call_batch)

        return length

    def enter_call(self):
        self._call_batches.append(OrderedDict())

    def revert_call(self):
        call_batch: OrderedDict = self._call_batches[-1]
        call_batch.clear()

    def leave_call(self):
        call_batch: OrderedDict = self._call_batches.pop()

        if call_batch:
            self._call_batches[-1].update(call_batch)

    def digest(self) -> bytes:
        if len(self._call_batches) != 1:
            raise DatabaseException(f'Wrong call_batch count: {len(self._call_batches)}')

        return digest(self._call_batches[0])

    @property
    def call_count(self) -> int:
        return len(self._call_batches)

    def clear(self):
        self.hash = None
        self._call_batches = [OrderedDict()]


class BlockBatch(Batch):
    """Contains the states changed by a block

    key: Address
    value: IconScoreBatch
    """

    def __init__(self, block: Optional['Block'] = None):
        """Constructor

        :param block: block info
        """
        super().__init__()
        self.block = block

    def __setitem__(self, key, value):
        raise AccessDeniedException("Can not set data on block batch directly.")

    def to_list(self) -> list:
        """
        Return list of key, value for Debugging
        :return:
        """
        new_list: list = []
        for key, val in self.items():
            _dict = {"key": key}
            _dict.update(val.to_dict(to_camel_case))
            new_list.append(_dict)
        return new_list

    def update(self, tx_batch: 'TransactionBatch', **kwargs):
        assert isinstance(tx_batch, TransactionBatch)

        for key, value in tx_batch.items():
            prev_block_batch_value: Optional['BlockBatchValue'] = self.get(key)
            if prev_block_batch_value is not None:
                tx_indexes: list = prev_block_batch_value.tx_indexes
                tx_indexes.append(value.tx_index)
            else:
                tx_indexes: list = [value.tx_index]
            bbv = BlockBatchValue(value.value, value.include_state_root_hash, tx_indexes)
            super().__setitem__(key, bbv)

    def update_block_hash(self, block_hash: bytes):
        self.block = Block(block_height=self.block.height,
                           block_hash=block_hash,
                           timestamp=self.block.timestamp,
                           prev_hash=self.block.prev_hash,
                           cumulative_fee=self.block.cumulative_fee)

    def set_block_to_batch(self, revision: int):
        # Logger.debug(tag="DB", msg=f"set_block_to_batch() block={self.block}")
        block_key: bytes = IcxStorage.LAST_BLOCK_KEY
        # As block is not relevant with a transaction, set tx_indexes as -1
        block_value: 'BlockBatchValue' = BlockBatchValue(self.block.to_bytes(revision), False, [-1])

        super().__setitem__(block_key, block_value)

    def clear(self) -> None:
        self.block = None
        super().clear()



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import TYPE_CHECKING, Optional, Tuple, Iterable

import plyvel
from iconcommons.logger import Logger

from .batch import TransactionBatchValue
from ..base.exception import DatabaseException
from ..icon_constant import ICON_DB_LOG_TAG, IconScoreContextType

if TYPE_CHECKING:
    from ..iconscore.icon_score_context import IconScoreContext


def _is_db_writable_on_context(context: 'IconScoreContext'):
    """Check if db is writable on a given context

    :param context:
    :return:
    """
    if context is None:
        return True
    else:
        return not context.readonly


class KeyValueDatabase(object):
    @staticmethod
    def from_path(path: str,
                  create_if_missing: bool = True) -> 'KeyValueDatabase':
        """

        :param path: db path
        :param create_if_missing:
        :return: KeyValueDatabase instance
        """
        db = plyvel.DB(path, create_if_missing=create_if_missing)
        return KeyValueDatabase(db)

    def __init__(self, db: plyvel.DB) -> None:
        """Constructor

        :param db: plyvel db instance
        """
        self._db = db

    def get(self, key: bytes) -> bytes:
        """Get the value for the specified key.

        :param key: (bytes): key to retrieve
        :return: value for the specified key, or None if not found
        """
        return self._db.get(key)

    def put(self, key: bytes, value: bytes) -> None:
        """Set a value for the specified key.

        :param key: (bytes): key to set
        :param value: (bytes): data to be stored
        """
        self._db.put(key, value)

    def delete(self, key: bytes) -> None:
        """Delete the key/value pair for the specified key.

        :param key: key to delete
        """
        self._db.delete(key)

    def close(self) -> None:
        """Close the database.
        """
        if self._db:
            self._db.close()
            self._db = None

    def get_sub_db(self, prefix: bytes) -> 'KeyValueDatabase':
        """Return a new prefixed database.

        :param prefix: (bytes): prefix to use
        """
        return KeyValueDatabase(self._db.prefixed_db(prefix))

    def iterator(self) -> iter:
        return self._db.iterator()

    def write_batch(self, it: Iterable[Tuple[bytes, Optional[bytes]]]) -> int:
        """Write a batch to the database for the specified states dict.

        :param it: iterable which return tuple(key, value)
            key: bytes
            value: optional bytes
        :return: the number of key-value pairs written
        """
        size = 0

        if it is None:
            return size

        with self._db.write_batch() as wb:
            for key, value in it:
                if value:
                    wb.put(key, value)
                else:
                    wb.delete(key)

                size += 1

        return size


class DatabaseObserver(object):
    """ An abstract class of database observer.
    """

    def __init__(self,
                 get_func: callable, put_func: callable, delete_func: callable):
        self.__get_func = get_func
        self.__put_func = put_func
        self.__delete_func = delete_func

    def on_get(self, context: 'IconScoreContext', key: bytes, value: bytes):
        """
        Invoked when `get` is called in `ContextDatabase`

        :param context: SCORE context
        :param key: key
        :param value: value
        """
        if not self.__get_func:
            Logger.warning('__get_func is None', ICON_DB_LOG_TAG)
        self.__get_func(context, key, value)

    def on_put(self,
               context: 'IconScoreContext',
               key: bytes,
               old_value: bytes,
               new_value: bytes):
        """Invoked when `put` is called in `ContextDatabase`.

        :param context: SCORE context
        :param key: key
        :param old_value: old value
        :param new_value: new value
        """
        if not self.__put_func:
            Logger.warning('__put_func is None', ICON_DB_LOG_TAG)
        self.__put_func(context, key, old_value, new_value)

    def on_delete(self,
                  context: 'IconScoreContext',
                  key: bytes,
                  old_value: bytes):
        """Invoked when `delete` is called in `ContextDatabase`.


        :param context: SCORE context
        :param key: key
        :param old_value:
        """
        if not self.__delete_func:
            Logger.warning('__delete_func is None', ICON_DB_LOG_TAG)
        self.__delete_func(context, key, old_value)


class ContextDatabase(object):
    """Database for an IconScore only used in the inside of iconservice.

    IconScore cannot access this database directly.
    Cache + LevelDB
    """

    def __init__(self, db: 'KeyValueDatabase', is_shared: bool = False) -> None:
        """Constructor

        :param db: KeyValueDatabase instance
        """
        self.key_value_db = db
        # True: this db is shared with all smart contracts
        self._is_shared = is_shared

    def get(self, context: Optional['IconScoreContext'], key: bytes) -> bytes:
        """Returns value indicated by key from batch or StateDB

        :param context:
        :param key:
        :return: value
        """
        context_type = context.type

        if context_type in (IconScoreContextType.DIRECT, IconScoreContextType.QUERY):
            return self.key_value_db.get(key)
        else:
            return self.get_from_batch(context, key)

    def get_from_batch(self,
                       context: 'IconScoreContext',
                       key: bytes) -> bytes:
        """Returns a value for a given key

        Search order
        1. TransactionBatch
        2. Current BlockBatch
        3. Prev BlockBatch
        4. StateDB

        :param context:
        :param key:

        :return: a value for a given key
        """
        # Find the value from tx_batch, block_batch and prev_block_batches with a given key
        for batch in context.get_batches():
            if key in batch:
                return batch[key].value

        # get value from state_db
        return self.key_value_db.get(key)

    @staticmethod
    def _check_tx_batch_value(context: Optional['IconScoreContext'],
                              key: bytes,
                              include_state_root_hash: bool):
        tx_batch_value: 'TransactionBatchValue' = context.tx_batch.get(key)
        if tx_batch_value is None:
            return

        if not isinstance(tx_batch_value, TransactionBatchValue):
            raise DatabaseException(
                f'Only TransactionBatchValue type is allowed on tx_batch: {type(tx_batch_value)}')
        elif tx_batch_value.include_state_root_hash != include_state_root_hash:
            raise DatabaseException('Do not change the include_state_root_hash on the same data')

    def put(self,
            context: Optional['IconScoreContext'],
            key: bytes,
            value: Optional[bytes]) -> None:
        """Set the value to StateDB or cache it according to context type

        :param context:
        :param key:
        :param value:
        """
        self._put(context, key, value, include_state_root_hash=True)

    def _put(self,
             context: Optional['IconScoreContext'],
             key: bytes,
             value: Optional[bytes],
             include_state_root_hash: bool) -> None:
        if not _is_db_writable_on_context(context):
            raise DatabaseException('No permission to write')

        context_type = context.type

        if context_type == IconScoreContextType.DIRECT:
            self.key_value_db.put(key, value)
        else:
            self._check_tx_batch_value(context, key, include_state_root_hash)
            tx_index: int = context.tx.index if context.tx is not None else -1
            context.tx_batch[key] = TransactionBatchValue(value, include_state_root_hash, tx_index)

    def delete(self,
               context: Optional['IconScoreContext'],
               key: bytes):
        """Delete key from db

        :param context:
        :param key: key to delete from db
        """
        self._delete(context, key, include_state_root_hash=True)

    def _delete(self,
                context: Optional['IconScoreContext'],
                key: bytes,
                include_state_root_hash: bool):
        if not _is_db_writable_on_context(context):
            raise DatabaseException('No permission to delete')

        context_type = context.type

        if context_type == IconScoreContextType.DIRECT:
            self.key_value_db.delete(key)
        else:
            self._check_tx_batch_value(context, key, include_state_root_hash)
            tx_index: int = context.tx.index if context.tx is not None else -1
            context.tx_batch[key] = TransactionBatchValue(None, include_state_root_hash, tx_index)

    def close(self, context: 'IconScoreContext') -> None:
        """close db

        :param context:
        """
        if not _is_db_writable_on_context(context):
            raise DatabaseException('No permission to close')

        if not self._is_shared:
            return self.key_value_db.close()

    def write_batch(self,
                    context: 'IconScoreContext',
                    it: Iterable[Tuple[bytes, Optional[bytes]]]):

        if not _is_db_writable_on_context(context):
            raise DatabaseException(
                'write_batch is not allowed on readonly context')

        return self.key_value_db.write_batch(it)

    @staticmethod
    def from_path(path: str,
                  create_if_missing: bool = True) -> 'ContextDatabase':
        db = KeyValueDatabase.from_path(path, create_if_missing)
        return ContextDatabase(db)


class MetaContextDatabase(ContextDatabase):
    def put(self,
            context: Optional['IconScoreContext'],
            key: bytes,
            value: Optional[bytes]) -> None:
        """Set the value to StateDB or cache it according to context type

        :param context:
        :param key:
        :param value:
        """
        self._put(context, key, value, include_state_root_hash=False)

    def delete(self,
               context: Optional['IconScoreContext'],
               key: bytes) -> None:
        """Set the value to StateDB or cache it according to context type

        :param context:
        :param key:
        """
        self._delete(context, key, include_state_root_hash=False)

    @staticmethod
    def from_path(path: str,
                  create_if_missing: bool = True) -> 'MetaContextDatabase':
        db = KeyValueDatabase.from_path(path, create_if_missing)
        return MetaContextDatabase(db)



# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..base.address import Address


def get_package_name_by_address_and_tx_hash(address: 'Address', tx_hash: bytes) -> str:
    return f'{address.to_bytes().hex()}.0x{tx_hash.hex()}'


def get_score_deploy_path(score_root_path: str, score_address: 'Address', tx_hash: bytes) -> str:
    return os.path.join(score_root_path, score_address.to_bytes().hex(), f'0x{tx_hash.hex()}')




def rlp_encode_bytes(b: bytes) -> bytes:
    blen = len(b)
    if blen == 1 and b[0] < 0x80:
        return b
    elif blen <= 55:
        return bytes([blen + 0x80]) + b
    len_bytes = rlp_get_bytes(blen)
    return bytes([len(len_bytes) + 0x80 + 55]) + len_bytes + b


def rlp_get_bytes(x: int) -> bytes:
    if x == 0:
        return b''
    else:
        return rlp_get_bytes(int(x / 256)) + bytes([x % 256])



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from enum import IntEnum

from .db import KeyValueDatabase, ContextDatabase
from ..base.address import Address
from ..icon_constant import ICON_DEX_DB_NAME


class ContextDatabaseFactory(object):

    class Mode(IntEnum):
        SINGLE_DB = 0
        MULTIPLE_DB = 1

    _state_db_root_path: str = None
    _mode: 'Mode' = Mode.SINGLE_DB
    _shared_context_db: 'ContextDatabase' = None

    @classmethod
    def open(cls, state_db_root_path: str, mode: 'Mode'):
        cls.close()

        cls._state_db_root_path = state_db_root_path
        cls._mode = mode

    @classmethod
    def get_shared_db(cls) -> ContextDatabase:
        if cls._shared_context_db is None:
            path = os.path.join(cls._state_db_root_path, ICON_DEX_DB_NAME)
            key_value_db = KeyValueDatabase.from_path(path)
            cls._shared_context_db = ContextDatabase(
                key_value_db, is_shared=True)

        return cls._shared_context_db

    @classmethod
    def create_by_address(cls, address: 'Address') -> ContextDatabase:
        if cls._mode == cls.Mode.SINGLE_DB:
            return cls.get_shared_db()
        else:
            return cls.create_by_name(address.body.hex())

    @classmethod
    def create_by_name(cls, name: str) -> ContextDatabase:
        if cls._mode == cls.Mode.SINGLE_DB:
            return cls.get_shared_db()
        else:
            path = os.path.join(cls._state_db_root_path, name)
            return ContextDatabase.from_path(path)

    @classmethod
    def close(cls):
        if cls._shared_context_db:
            cls._shared_context_db.key_value_db.close()
            cls._shared_context_db = None



# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Any, TYPE_CHECKING, Optional

from earlgrey import message_queue_task, MessageQueueStub, MessageQueueService

from iconcommons.logger import Logger
from iconservice.base.address import Address
from iconservice.base.block import Block
from iconservice.base.exception import ExceptionCode, IconServiceBaseException, InvalidBaseTransactionException, \
    FatalException, ServiceNotReadyException
from iconservice.base.type_converter import TypeConverter, ParamType
from iconservice.base.type_converter_templates import ConstantKeys
from iconservice.icon_constant import EnableThreadFlag, ENABLE_THREAD_FLAG, RPCMethod
from iconservice.icon_service_engine import IconServiceEngine
from iconservice.utils import check_error_response, to_camel_case, BytesToHexJSONEncoder, bytes_to_hex

if TYPE_CHECKING:
    from earlgrey import RobustConnection

THREAD_INVOKE = 'invoke'
THREAD_QUERY = 'query'
THREAD_ESTIMATE = 'estimate'
THREAD_VALIDATE = 'validate'
THREAD_STATUS = 'status'

QUERY_THREAD_MAPPER = {
    RPCMethod.ICX_GET_BALANCE: THREAD_STATUS,
    RPCMethod.ICX_GET_TOTAL_SUPPLY: THREAD_STATUS,
    RPCMethod.ICX_GET_SCORE_API: THREAD_STATUS,
    RPCMethod.ISE_GET_STATUS: THREAD_STATUS,
    RPCMethod.ICX_CALL: THREAD_QUERY,
    RPCMethod.DEBUG_ESTIMATE_STEP: THREAD_ESTIMATE,
    RPCMethod.DEBUG_GET_ACCOUNT: THREAD_QUERY,
}

_TAG = "MQ"


class IconScoreInnerTask(object):
    def __init__(self, conf: dict):
        self._conf = conf
        self._thread_flag = ENABLE_THREAD_FLAG

        self._icon_service_engine = IconServiceEngine()
        self._open()

        self._thread_pool = {
            THREAD_INVOKE: ThreadPoolExecutor(1),
            THREAD_STATUS: ThreadPoolExecutor(1),
            THREAD_QUERY: ThreadPoolExecutor(1),
            THREAD_ESTIMATE: ThreadPoolExecutor(1),
            THREAD_VALIDATE: ThreadPoolExecutor(1)
        }

    def _open(self):
        Logger.info(tag=_TAG, msg="_open() start")
        self._icon_service_engine.open(self._conf)
        Logger.info(tag=_TAG, msg="_open() end")

    def _is_thread_flag_on(self, flag: 'EnableThreadFlag') -> bool:
        return (self._thread_flag & flag) == flag

    def _check_icon_service_ready(self):
        if not self._icon_service_engine.is_reward_calculator_ready():
            raise ServiceNotReadyException("Reward Calculator is not ready")

    @staticmethod
    def _log_exception(e: BaseException, tag: str) -> None:
        Logger.exception(str(e), tag)
        Logger.error(str(e), tag)

    @message_queue_task
    async def hello(self):
        Logger.info(tag=_TAG, msg='hello() start')

        ready_future = self._icon_service_engine.get_ready_future()
        await ready_future

        if self._is_thread_flag_on(EnableThreadFlag.INVOKE):
            loop = asyncio.get_event_loop()
            ret = await loop.run_in_executor(self._thread_pool[THREAD_INVOKE], self._hello)
        else:
            ret = self._hello()

        Logger.info(tag=_TAG, msg='hello() end')

        return ret

    def _hello(self):
        return self._icon_service_engine.hello()

    def cleanup(self):
        Logger.info(tag=_TAG, msg="cleanup() start")

        # shutdown thread pool executors
        for executor in self._thread_pool.values():
            executor.shutdown()

        # close ICON Service
        if self._icon_service_engine:
            self._icon_service_engine.close()
            self._icon_service_engine = None

        Logger.info(tag=_TAG, msg="cleanup() end")

    @message_queue_task
    async def close(self):
        Logger.info(tag=_TAG, msg="close() stop event loop")
        self._close()

    @staticmethod
    def _close():
        asyncio.get_event_loop().stop()

    @message_queue_task
    async def invoke(self, request: dict) -> dict:
        Logger.debug(tag=_TAG, msg=f'invoke() start')

        self._check_icon_service_ready()

        if self._is_thread_flag_on(EnableThreadFlag.INVOKE):
            loop = asyncio.get_event_loop()
            ret: dict = await loop.run_in_executor(self._thread_pool[THREAD_INVOKE],
                                                   self._invoke, request)
        else:
            ret: dict = self._invoke(request)

        Logger.debug(tag=_TAG, msg=f'invoke() end')
        return ret

    def _invoke(self, request: dict):
        """Process transactions in a block

        :param request:
        :return:
        """

        Logger.info(tag=_TAG, msg=f'INVOKE Request: {request}')

        try:
            params = TypeConverter.convert(request, ParamType.INVOKE)
            converted_block_params = params['block']
            block = Block.from_dict(converted_block_params)
            Logger.info(tag=_TAG, msg=f'INVOKE: BH={block.height}')

            converted_tx_requests = params['transactions']

            convert_tx_result_to_dict: bool = 'isBlockEditable' in params

            converted_is_block_editable = params.get('isBlockEditable', False)
            converted_prev_block_generator = params.get('prevBlockGenerator')
            converted_prev_block_validators = params.get('prevBlockValidators')
            converted_prev_votes = params.get('prevBlockVotes')

            tx_results, state_root_hash, added_transactions, next_preps = self._icon_service_engine.invoke(
                block=block,
                tx_requests=converted_tx_requests,
                prev_block_generator=converted_prev_block_generator,
                prev_block_validators=converted_prev_block_validators,
                prev_block_votes=converted_prev_votes,
                is_block_editable=converted_is_block_editable)

            if convert_tx_result_to_dict:
                convert_tx_results = [tx_result.to_dict(to_camel_case) for tx_result in tx_results]
            else:
                # old version
                convert_tx_results = {bytes.hex(tx_result.tx_hash): tx_result.to_dict(to_camel_case)
                                      for tx_result in tx_results}
            results = {
                'txResults': convert_tx_results,
                'stateRootHash': bytes.hex(state_root_hash),
                'addedTransactions': added_transactions
            }

            if next_preps:
                results["prep"] = next_preps

            response = MakeResponse.make_response(results)
        except FatalException as e:
            self._log_exception(e, _TAG)
            response = MakeResponse.make_error_response(ExceptionCode.SYSTEM_ERROR, str(e))
            self._close()
        except InvalidBaseTransactionException as e:
            self._log_exception(e, _TAG)
            response = MakeResponse.make_error_response(ExceptionCode.SYSTEM_ERROR, str(e))
        except IconServiceBaseException as icon_e:
            self._log_exception(icon_e, _TAG)
            response = MakeResponse.make_error_response(icon_e.code, icon_e.message)
        except Exception as e:
            self._log_exception(e, _TAG)
            response = MakeResponse.make_error_response(ExceptionCode.SYSTEM_ERROR, str(e))
        finally:
            if self._icon_service_engine:
                self._icon_service_engine.clear_context_stack()

        Logger.info(tag=_TAG, msg=f'INVOKE Response: {json.dumps(response, cls=BytesToHexJSONEncoder)}')
        return response

    @message_queue_task
    async def query(self, request: dict) -> dict:
        self._check_icon_service_ready()
        return await self._get_query_response(request)

    async def _get_query_response(self, request: dict) -> dict:
        try:
            value = await self._execute_query(request)
            if isinstance(value, Address):
                value = str(value)
            response = MakeResponse.make_response(value)
        except FatalException as e:
            self._log_exception(e, _TAG)
            response = MakeResponse.make_error_response(ExceptionCode.SYSTEM_ERROR, str(e))
        except IconServiceBaseException as icon_e:
            self._log_exception(icon_e, _TAG)
            response = MakeResponse.make_error_response(icon_e.code, icon_e.message)
        except Exception as e:
            self._log_exception(e, _TAG)
            response = MakeResponse.make_error_response(ExceptionCode.SYSTEM_ERROR, str(e))

        self._icon_service_engine.clear_context_stack()
        return response

    async def _execute_query(self, request: dict):
        method_name: str = request['method']
        if method_name == RPCMethod.DEBUG_ESTIMATE_STEP:
            method: callable = self._estimate
            args = [request]
        else:
            method: callable = self._query
            args = [request, method_name]

        if self._is_thread_flag_on(EnableThreadFlag.QUERY):
            return await asyncio.get_event_loop(). \
                run_in_executor(self._thread_pool[QUERY_THREAD_MAPPER[method_name]],
                                method, *args)
        else:
            return method(*args)

    def _estimate(self, request: dict):
        converted_request = TypeConverter.convert(request, ParamType.INVOKE_TRANSACTION)
        return self._icon_service_engine.estimate_step(converted_request)

    def _query(self, request: dict, method: str):
        converted_request = TypeConverter.convert(request, ParamType.QUERY)
        return self._icon_service_engine.query(method, converted_request['params'])

    @message_queue_task
    async def call(self, request: dict):
        Logger.info(tag=_TAG, msg=f'call() start: {request}')

        self._check_icon_service_ready()

        if self._is_thread_flag_on(EnableThreadFlag.QUERY):
            loop = asyncio.get_event_loop()
            ret = await loop.run_in_executor(self._thread_pool[THREAD_QUERY],
                                             self._call, request)
        else:
            ret = self._call(request)

        Logger.info(tag=_TAG, msg=f'call() end: {ret}')
        return ret

    def _call(self, request: dict):
        try:
            response = self._icon_service_engine.inner_call(request)

            if isinstance(response, Address):
                response = str(response)
        except FatalException as e:
            self._log_exception(e, _TAG)
            response = MakeResponse.make_error_response(ExceptionCode.SYSTEM_ERROR, str(e))
        except IconServiceBaseException as icon_e:
            self._log_exception(icon_e, _TAG)
            response = MakeResponse.make_error_response(icon_e.code, icon_e.message)
        except Exception as e:
            self._log_exception(e, _TAG)
            response = MakeResponse.make_error_response(ExceptionCode.SYSTEM_ERROR, str(e))

        return response

    @message_queue_task
    async def write_precommit_state(self, request: dict):
        self._check_icon_service_ready()

        if self._is_thread_flag_on(EnableThreadFlag.INVOKE):
            loop = asyncio.get_event_loop()
            ret = await loop.run_in_executor(self._thread_pool[THREAD_INVOKE],
                                             self._write_precommit_state, request)
        else:
            ret = self._write_precommit_state(request)

        return ret

    def _write_precommit_state(self, request: dict) -> dict:
        Logger.info(tag=_TAG, msg=f'WRITE_PRECOMMIT_STATE Request: {request}')

        try:
            converted_params = TypeConverter.convert(request, ParamType.WRITE_PRECOMMIT)
            block_height: int = converted_params[ConstantKeys.BLOCK_HEIGHT]
            instant_block_hash: bytes = converted_params[ConstantKeys.OLD_BLOCK_HASH]
            block_hash = converted_params[ConstantKeys.NEW_BLOCK_HASH]

            Logger.info(tag=_TAG, msg=f'WRITE_PRECOMMIT_STATE: '
                                      f'BH={block_height} '
                                      f'instant_block_hash={bytes_to_hex(instant_block_hash)} '
                                      f'block_hash={bytes_to_hex(block_hash)}')

            self._icon_service_engine.commit(block_height, instant_block_hash, block_hash)
            response = MakeResponse.make_response(ExceptionCode.OK)
        except FatalException as e:
            self._log_exception(e, _TAG)
            response = MakeResponse.make_error_response(ExceptionCode.SYSTEM_ERROR, str(e))
            self._close()
        except IconServiceBaseException as icon_e:
            self._log_exception(icon_e, _TAG)
            response = MakeResponse.make_error_response(icon_e.code, icon_e.message)
        except Exception as e:
            self._log_exception(e, _TAG)
            response = MakeResponse.make_error_response(ExceptionCode.SYSTEM_ERROR, str(e))

        Logger.info(tag=_TAG, msg=f'WRITE_PRECOMMIT_STATE Response: {response}')
        return response

    @message_queue_task
    async def remove_precommit_state(self, request: dict):
        Logger.info(tag=_TAG, msg=f'remove_precommit_state() start')

        self._check_icon_service_ready()


        """
        Unused API
        """
        return {}

    @message_queue_task
    async def rollback(self, request: dict):
        """Go back to the state of the given previous block

        :param request:
        :return:
        """

        Logger.info(tag=_TAG, msg=f"rollback() start")

        self._check_icon_service_ready()

        if self._is_thread_flag_on(EnableThreadFlag.INVOKE):
            loop = asyncio.get_event_loop()
            response = await loop.run_in_executor(self._thread_pool[THREAD_INVOKE], self._rollback, request)
        else:
            response = self._rollback(request)

        Logger.info(tag=_TAG, msg=f"rollback() end")

        return response

    def _rollback(self, request: dict) -> dict:
        Logger.info(tag=_TAG, msg=f"ROLLBACK Request: {request}")

        try:
            converted_params = TypeConverter.convert(request, ParamType.ROLLBACK)
            block_height: int = converted_params[ConstantKeys.BLOCK_HEIGHT]
            block_hash: bytes = converted_params[ConstantKeys.BLOCK_HASH]
            Logger.info(tag=_TAG, msg=f"ROLLBACK: BH={block_height} block_hash={bytes_to_hex(block_hash)}")

            response: dict = self._icon_service_engine.rollback(block_height, block_hash)
            response = MakeResponse.make_response(response)
        except FatalException as e:
            self._log_exception(e, _TAG)
            response = MakeResponse.make_error_response(ExceptionCode.SYSTEM_ERROR, str(e))
            self._close()
        except IconServiceBaseException as icon_e:
            self._log_exception(icon_e, _TAG)
            response = MakeResponse.make_error_response(icon_e.code, icon_e.message)
        except BaseException as e:
            self._log_exception(e, _TAG)
            response = MakeResponse.make_error_response(ExceptionCode.SYSTEM_ERROR, str(e))

        Logger.info(tag=_TAG, msg=f"ROLLBACK Response: {response}")
        return response

    @message_queue_task
    async def validate_transaction(self, request: dict):
        self._check_icon_service_ready()

        if self._is_thread_flag_on(EnableThreadFlag.VALIDATE):
            loop = asyncio.get_event_loop()
            return await loop.run_in_executor(self._thread_pool[THREAD_VALIDATE],
                                              self._validate_transaction, request)
        else:
            return self._validate_transaction(request)

    def _validate_transaction(self, request: dict):
        try:
            Logger.info(tag=_TAG, msg=f'validate_transaction Request: {request}')
            converted_request = TypeConverter.convert(request, ParamType.VALIDATE_TRANSACTION)
            self._icon_service_engine.validate_transaction(converted_request, request)
            response = MakeResponse.make_response(ExceptionCode.OK)
        except FatalException as e:
            self._log_exception(e, _TAG)
            response = MakeResponse.make_error_response(ExceptionCode.SYSTEM_ERROR, str(e))
        except IconServiceBaseException as icon_e:
            self._log_exception(icon_e, _TAG)
            response = MakeResponse.make_error_response(icon_e.code, icon_e.message)
        except Exception as e:
            self._log_exception(e, _TAG)
            response = MakeResponse.make_error_response(ExceptionCode.SYSTEM_ERROR, str(e))

        self._icon_service_engine.clear_context_stack()
        return response

    @message_queue_task
    async def change_block_hash(self, _params):
        self._check_icon_service_ready()
        return ExceptionCode.OK


class MakeResponse:
    @staticmethod
    def make_response(response: Any):
        if check_error_response(response):
            return response
        else:
            return TypeConverter.convert_type_reverse(response)

    @staticmethod
    def make_error_response(code: Any, message: str) -> dict:
        _code: int = int(code) + 32000
        return {'error': {'code': _code, 'message': message}}


class IconScoreInnerService(MessageQueueService[IconScoreInnerTask]):
    TaskType = IconScoreInnerTask

    def _callback_connection_close(self, sender, exc: Optional[BaseException], *args, **kwargs):
        Logger.error(tag=_TAG, msg=f"[Inner Service] connection closed. {exc}")
        self.clean_close()

    def clean_close(self):
        Logger.debug(tag=_TAG, msg="icon service will be closed")
        self._task.cleanup()


class IconScoreInnerStub(MessageQueueStub[IconScoreInnerTask]):
    TaskType = IconScoreInnerTask

    def _callback_connection_close(self, sender, exc: Optional[BaseException], *args, **kwargs):
        Logger.error(tag=_TAG, msg=f"[Inner Stub] connection closed. {exc}")



# -*- coding: utf-8 -*-
# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Optional

from iconcommons import Logger

from .issue_formula import IssueFormula
from .regulator import Regulator
from ... import SYSTEM_SCORE_ADDRESS
from ...base.ComponentBase import EngineBase
from ...base.exception import (
    OutOfBalanceException,
    InvalidParamsException,
    InternalServiceErrorException,
)
from ...icon_constant import ISSUE_CALCULATE_ORDER, ISSUE_EVENT_LOG_MAPPER, IssueDataKey, ICX_LOG_TAG, Revision
from ...iconscore.icon_score_event_log import EventLogEmitter

if TYPE_CHECKING:
    from ...iconscore.icon_score_context import IconScoreContext
    from ...base.address import Address
    from ...icx.icx_account import Account


class Engine(EngineBase):

    def __init__(self):
        super().__init__()

        self._formula: Optional['IssueFormula'] = None

    def open(self, context: 'IconScoreContext'):
        self._formula = IssueFormula(context.main_prep_count)

    def create_icx_issue_info(self, context: 'IconScoreContext') -> dict:
        irep: int = context.engine.prep.term.irep
        iiss_data_for_issue = {
            IssueDataKey.PREP: {
                IssueDataKey.IREP: irep,
                IssueDataKey.RREP: context.storage.iiss.get_reward_rate(context).reward_prep,
                IssueDataKey.TOTAL_DELEGATION: context.preps.total_delegated
            }
        }
        total_issue_amount = 0
        for group in iiss_data_for_issue:
            issue_amount_per_group = self._formula.calculate(context, group, iiss_data_for_issue[group])
            iiss_data_for_issue[group][IssueDataKey.VALUE] = issue_amount_per_group
            total_issue_amount += issue_amount_per_group

        context.regulator = Regulator(context, total_issue_amount)

        iiss_data_for_issue[IssueDataKey.ISSUE_RESULT] = {
            IssueDataKey.COVERED_BY_FEE: context.regulator.covered_icx_by_fee,
            IssueDataKey.COVERED_BY_OVER_ISSUED_ICX: context.regulator.covered_icx_by_over_issue,
            IssueDataKey.ISSUE: context.regulator.corrected_icx_issue_amount
        }
        return iiss_data_for_issue

    @staticmethod
    def _issue(context: 'IconScoreContext',
               to: 'Address',
               amount: int):
        if amount > 0:
            to_account: 'Account' = context.storage.icx.get_account(context, to)
            to_account.deposit(amount)
            current_total_supply = context.storage.icx.get_total_supply(context)
            context.storage.icx.put_account(context, to_account)
            context.storage.icx.put_total_supply(context, current_total_supply + amount)
            Logger.info(f"Issue icx. amount: {amount} "
                        f"Total supply: {current_total_supply + amount} "
                        f"Treasury: {to_account.balance}", ICX_LOG_TAG)

    def issue(self,
              context: 'IconScoreContext',
              to_address: 'Address',
              issue_data: dict):
        assert isinstance(context.regulator, Regulator)

        self._issue(context, to_address, context.regulator.corrected_icx_issue_amount)
        context.regulator.put_regulate_variable(context)

        for group_key in ISSUE_CALCULATE_ORDER:
            if group_key not in issue_data:
                continue
            event_signature: str = ISSUE_EVENT_LOG_MAPPER[group_key]["event_signature"]
            data: list = [issue_data[group_key][data_key] for data_key in ISSUE_EVENT_LOG_MAPPER[group_key]["data"]]
            EventLogEmitter.emit_event_log(context,
                                           score_address=SYSTEM_SCORE_ADDRESS,
                                           event_signature=event_signature,
                                           arguments=data,
                                           indexed_args_count=0)

        EventLogEmitter.emit_event_log(context,
                                       score_address=SYSTEM_SCORE_ADDRESS,
                                       event_signature=ISSUE_EVENT_LOG_MAPPER[IssueDataKey.TOTAL]["event_signature"],
                                       arguments=[context.regulator.covered_icx_by_fee,
                                                  context.regulator.covered_icx_by_over_issue,
                                                  context.regulator.corrected_icx_issue_amount,
                                                  context.regulator.remain_over_issued_icx],
                                       indexed_args_count=0)

    @staticmethod
    def _burn(context: 'IconScoreContext', address: 'Address', amount: int) -> int:
        if context.revision >= Revision.BURN_V2_ENABLED.value:
            address = SYSTEM_SCORE_ADDRESS

        account: 'Account' = context.storage.icx.get_account(context, address)
        if account.balance < amount:
            raise OutOfBalanceException(
                f'Not enough icx to burn: '
                f'balance({account.balance}) < icx_to_burn({amount})'
            )

        old_total_supply: int = context.storage.icx.get_total_supply(context)
        new_total_supply = old_total_supply - amount
        if new_total_supply < 0:
            raise InternalServiceErrorException(
                "Failed to burn icx: "
                f"total_supply({old_total_supply}) < icx_to_burn({amount})"
            )

        account.withdraw(amount)
        context.storage.icx.put_account(context, account)
        context.storage.icx.put_total_supply(context, new_total_supply)

        # Verify whether total_supply decreased by amount
        total_supply: int = context.storage.icx.get_total_supply(context)
        if total_supply != new_total_supply:
            raise InternalServiceErrorException(
                f"Failed to burn icx: new_total_supply={total_supply} "
                f"old_total_supply={old_total_supply} "
                f"icx_to_burn={amount}"
            )

        return new_total_supply

    def burn(self, context: 'IconScoreContext', address: 'Address', amount: int):
        """

        :param context:
        :param address: The address to burn ICX
        :param amount: the amount of ICX to burn
        :return:
        """
        revision: int = context.revision

        if revision >= Revision.BURN_V2_ENABLED.value and amount <= 0:
            raise InvalidParamsException(f"Invalid amount: {amount}")

        new_total_supply: int = self._burn(context, address, amount)
        self._log_burn_event(context, address, amount, new_total_supply)

    @staticmethod
    def _log_burn_event(
            context: 'IconScoreContext', address: 'Address', amount: int, new_total_supply: int):
        revision = context.revision

        # Event signature
        if revision < Revision.FIX_BURN_EVENT_SIGNATURE.value:
            event_sig = "ICXBurned"
        elif revision < Revision.BURN_V2_ENABLED.value:
            event_sig = "ICXBurned(int)"
        else:
            # 0: Who burns ICX
            # 1: The amount of ICX to burn
            # 2: New total supply
            event_sig = "ICXBurnedV2(Address,int,int)"

        # Arguments
        if revision < Revision.BURN_V2_ENABLED.value:
            arguments = [amount]
            indexed_args_count = 0
        else:
            arguments = [address, amount, new_total_supply]
            indexed_args_count = 1

        # Log event
        EventLogEmitter.emit_event_log(
            context,
            score_address=SYSTEM_SCORE_ADDRESS,
            event_signature=event_sig,
            arguments=arguments,
            indexed_args_count=indexed_args_count
        )



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .address import Address


class Transaction(object):
    """
    Holds information of the transaction
    """

    def __init__(self,
                 tx_hash: Optional[bytes] = None,
                 index: int = 0,
                 origin: Optional['Address'] = None,
                 to: Optional['Address'] = None,
                 timestamp: int = None,
                 nonce: int = None) -> None:
        """Transaction class for icon score context
        """
        self._hash = tx_hash
        self._index = index
        self._origin = origin
        self._to = to
        self._timestamp = timestamp
        self._nonce = nonce

    @property
    def origin(self) -> 'Address':
        """
        The account who created the transaction.
        """
        return self._origin

    @property
    def index(self) -> int:
        """
        Transaction index in a block
        """
        return self._index

    @property
    def hash(self) -> bytes:
        """
        Transaction hash
        """
        return self._hash

    @property
    def timestamp(self) -> int:
        """
        Timestamp of a transaction request in microseconds
        This is NOT a block timestamp
        """
        return self._timestamp

    @property
    def nonce(self) -> int:
        """
        (optional)
        nonce of a transaction request.
        random value
        """
        return self._nonce

    @property
    def to(self) -> 'Address':
        """
        The account of tx to.
        """
        return self._to




# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from abc import abstractmethod, ABC, ABCMeta
from functools import partial, wraps
from inspect import isfunction, signature, Parameter
from typing import TYPE_CHECKING, Callable, Any, List, Tuple, Mapping

from .context.context import ContextGetter, ContextContainer
from .db import IconScoreDatabase
from .icon_score_base2 import InterfaceScore, revert, Block
from .icon_score_constant import (
    CONST_INDEXED_ARGS_COUNT,
    FORMAT_IS_NOT_FUNCTION_OBJECT,
    ScoreFlag,
    FORMAT_DECORATOR_DUPLICATED,
    FORMAT_IS_NOT_DERIVED_OF_OBJECT,
    STR_FALLBACK,
    CONST_CLASS_API,
    CONST_CLASS_ELEMENT_METADATAS,
    BaseType,
    T,
)
from .icon_score_context_util import IconScoreContextUtil
from .icon_score_event_log import EventLogEmitter
from .icon_score_step import StepType
from .icx import Icx
from .internal_call import InternalCall
from .typing.definition import get_score_api
from .typing.element import (
    ScoreElementMetadataContainer,
    ScoreElementMetadata,
    FunctionMetadata,
    set_score_flag_on,
    is_any_score_flag_on,
)
from .typing.element import create_score_element_metadatas
from ..base.address import Address
from ..base.address import GOVERNANCE_SCORE_ADDRESS
from ..base.exception import *
from ..database.db import DatabaseObserver
from ..icon_constant import ICX_TRANSFER_EVENT_LOG, Revision, IconScoreContextType
from ..utils import get_main_type_from_annotations_type

if TYPE_CHECKING:
    from .icon_score_context import IconScoreContext
    from ..base.transaction import Transaction
    from ..base.message import Message

INDEXED_ARGS_LIMIT = 3


def interface(func):
    """
    A decorator for the functions of InterfaceScore.

    If other SCORE has the function whose signature is the same as defined with @interface decorator,
    the function can be invoked via InterfaceScore class instance
    """

    cls_name, func_name = str(func.__qualname__).split('.')
    if not isfunction(func):
        raise IllegalFormatException(FORMAT_IS_NOT_FUNCTION_OBJECT.format(func, cls_name))

    if is_any_score_flag_on(func, ScoreFlag.INTERFACE):
        raise InvalidInterfaceException(FORMAT_DECORATOR_DUPLICATED.format('interface', func_name, cls_name))

    set_score_flag_on(func, ScoreFlag.INTERFACE)

    @wraps(func)
    def __wrapper(calling_obj: "InterfaceScore", *args, **kwargs):
        if not isinstance(calling_obj, InterfaceScore):
            raise InvalidInstanceException(
                FORMAT_IS_NOT_DERIVED_OF_OBJECT.format(InterfaceScore.__name__))

        context = ContextContainer._get_context()
        addr_to = calling_obj.addr_to
        addr_from: 'Address' = context.current_address

        amount: int = getattr(calling_obj, "_InterfaceScore__get_icx")()
        getattr(calling_obj, "_InterfaceScore__reset_icx")()

        if addr_to is None:
            raise InvalidInterfaceException('Cannot create an interface SCORE with a None address')

        return InternalCall.other_external_call(context, addr_from, addr_to, amount, func_name, args, kwargs)

    return __wrapper


def eventlog(func=None, *, indexed=0):
    """
    Functions with `@eventlog` decorator will include logs in its TxResult as ‘eventlogs’.
    If indexed parameter is set in the decorator, designated number of parameters in the order of
    declaration will be indexed and included in the Bloom filter.
    Indexed parameters and non-indexed parameters are separately stored in TxResult.
    Possible data types for function parameters are primitive types (int, str, bytes, bool, Address).

    It is recommended to declare a function without implementation body.
    Even if the function has a body, it does not be executed.
    When declaring a function, type hinting is a must. Without type hinting, transaction will fail.
    The default value for the parameter can be set.
    At most 3 parameters can be indexed, And index can’t exceed the number of parameters(will raise an error).

    :param indexed: the number of indexed parameters count(maximum 3)
    """
    if func is None:
        return partial(eventlog, indexed=indexed)

    cls_name, func_name = str(func.__qualname__).split('.')
    if not isfunction(func):
        raise IllegalFormatException(FORMAT_IS_NOT_FUNCTION_OBJECT.format(func, cls_name))

    if not list(signature(func).parameters.keys())[0] == 'self':
        raise InvalidEventLogException("'self' is not declared as the first parameter")
    if indexed > INDEXED_ARGS_LIMIT:
        raise InvalidEventLogException(
            f'Indexed arguments overflow: limit={INDEXED_ARGS_LIMIT}')

    parameters = signature(func).parameters.values()
    if len(parameters) - 1 < indexed:
        raise InvalidEventLogException("Index exceeds the number of parameters")

    if is_any_score_flag_on(func, ScoreFlag.EVENTLOG):
        raise InvalidEventLogException(FORMAT_DECORATOR_DUPLICATED.format('eventlog', func_name, cls_name))

    set_score_flag_on(func, ScoreFlag.EVENTLOG)
    setattr(func, CONST_INDEXED_ARGS_COUNT, indexed)
    event_signature = __retrieve_event_signature(func_name, parameters)

    @wraps(func)
    def __wrapper(calling_obj: 'IconScoreBase', *args, **kwargs):
        if not (isinstance(calling_obj, IconScoreBase)):
            raise InvalidInstanceException(
                FORMAT_IS_NOT_DERIVED_OF_OBJECT.format(IconScoreBase.__name__))
        try:
            arguments = __resolve_arguments(func_name, parameters, args, kwargs)
        except IllegalFormatException as e:
            raise InvalidEventLogException(e.message)

        if event_signature == ICX_TRANSFER_EVENT_LOG:
            # 'ICXTransfer(Address,Address,int)' is reserved
            raise InvalidEventLogException(
                f'The event log \'{ICX_TRANSFER_EVENT_LOG}\' is reserved')

        return EventLogEmitter.emit_event_log(context=calling_obj._context,
                                              score_address=calling_obj.address,
                                              event_signature=event_signature,
                                              arguments=arguments,
                                              indexed_args_count=indexed,
                                              fee_charge=True)

    return __wrapper


def __retrieve_event_signature(function_name, parameters) -> str:
    """
    Retrieves a event signature from the function name and parameters
    :param function_name: name of event function
    :param parameters: Arguments description of the function declaration
    :return: event signature
    """
    type_names: List[str] = []
    for i, param in enumerate(parameters):
        if i > 0:
            # If there's no hint of argument in the function declaration,
            # raise an exception
            if param.annotation is Parameter.empty:
                raise IllegalFormatException(
                    f"Missing argument hint for '{function_name}': '{param.name}'")

            main_type = None
            if isinstance(param.annotation, type):
                main_type = param.annotation
            elif param.annotation == 'Address':
                main_type = Address

            # Raises an exception if the types are not supported
            if main_type is None or not issubclass(main_type, BaseType.__constraints__):
                raise IllegalFormatException(
                    f"Unsupported type for '{param.name}: {param.annotation}'")

            type_names.append(str(main_type.__name__))
    return f"{function_name}({','.join(type_names)})"


def __resolve_arguments(function_name, parameters, args, kwargs) -> List[Any]:
    """
    Resolves arguments with keeping order as the function declaration
    :param parameters: Arguments description of the function declaration
    :param args: input ordered arguments
    :param kwargs: input keyword arguments
    :return: an ordered list of arguments
    """
    arguments = []
    if len(parameters) - 1 < len(args) + len(kwargs):
        raise IllegalFormatException(
            f"The maximum number of arguments which event log method({function_name}) can accept is exceeded")

    for i, parameter in enumerate(parameters, -1):
        if i < 0:
            # pass the self parameter
            continue
        name = parameter.name
        annotation = parameter.annotation
        if i < len(args):
            # the argument is in the ordered args
            value = args[i]
            if name in kwargs:
                raise IllegalFormatException(
                    f"Duplicated argument value for '{function_name}': {name}")
        else:
            # If arg is over, the argument should be searched on kwargs
            try:
                value = kwargs[name]
            except KeyError:
                if not parameter.default == Parameter.empty:
                    value = parameter.default
                else:
                    raise IllegalFormatException(
                        f"Missing argument value for '{function_name}': {name}")

        main_type = get_main_type_from_annotations_type(annotation)

        if main_type == 'Address':
            main_type = Address

        if value is not None and not isinstance(value, main_type):
            raise IllegalFormatException(
                f"Type mismatch of '{name}': {type(value)}, expected: {main_type}")
        arguments.append(value)
    return arguments


def external(func=None, *, readonly=False):
    """
    A decorator for the function whether the function exposes externally.
    If declared to the function, EOA or another SCORE can call it.
    These functions are registered on the exportable API list.
    Any attempt to call a non-external function from outside the contract will fail.

    If a function is decorated with ‘readonly’ parameters, i.e., `@external(readonly=True)`,
    the function will have read-only access to the state DB. This is similar to view keyword in Solidity.
    If the read-only external function is also decorated with `@payable`, the function call will fail.
    Duplicate declaration of @external will raise an exception on import time.

    :param readonly: True if the function have read-only access to the state DB.
    """
    if func is None:
        return partial(external, readonly=readonly)

    cls_name, func_name = str(func.__qualname__).split('.')
    if not isfunction(func):
        raise IllegalFormatException(FORMAT_IS_NOT_FUNCTION_OBJECT.format(func, cls_name))

    if func_name == STR_FALLBACK:
        raise InvalidExternalException(f"{func_name} cannot be declared as external")

    if is_any_score_flag_on(func, ScoreFlag.EXTERNAL):
        raise InvalidExternalException(FORMAT_DECORATOR_DUPLICATED.format('external', func_name, cls_name))

    score_flag = ScoreFlag.EXTERNAL
    if readonly:
        score_flag |= ScoreFlag.READONLY
    set_score_flag_on(func, score_flag)

    @wraps(func)
    def __wrapper(calling_obj: Any, *args, **kwargs):
        if not (isinstance(calling_obj, IconScoreBase)):
            raise InvalidInstanceException(
                FORMAT_IS_NOT_DERIVED_OF_OBJECT.format(IconScoreBase.__name__))
        res = func(calling_obj, *args, **kwargs)
        return res

    return __wrapper


def payable(func):
    """
    A decorator for the external function.

    If the decorator is declared to the external function,
    it can receive the ICXs and process further works for it.
    If ICXs (msg.value) are passed to a non-payable function, that transaction will fail.
    """
    cls_name, func_name = str(func.__qualname__).split('.')
    if not isfunction(func):
        raise IllegalFormatException(FORMAT_IS_NOT_FUNCTION_OBJECT.format(func, cls_name))

    if is_any_score_flag_on(func, ScoreFlag.PAYABLE):
        raise InvalidPayableException(FORMAT_DECORATOR_DUPLICATED.format('payable', func_name, cls_name))

    flag = ScoreFlag.PAYABLE
    if func_name == STR_FALLBACK:
        # If a function has payable decorator and its name is "fallback",
        # then it is a fallback function
        flag |= ScoreFlag.FALLBACK

    set_score_flag_on(func, flag)

    @wraps(func)
    def __wrapper(calling_obj: Any, *args, **kwargs):
        if not (isinstance(calling_obj, IconScoreBase)):
            raise InvalidInstanceException(
                FORMAT_IS_NOT_DERIVED_OF_OBJECT.format(IconScoreBase.__name__))
        res = func(calling_obj, *args, **kwargs)
        return res

    return __wrapper


class IconScoreObject(ABC):

    def __init__(self, *args, **kwargs) -> None:
        pass

    def on_install(self, **kwargs) -> None:
        pass

    def on_update(self, **kwargs) -> None:
        pass


class IconScoreBaseMeta(ABCMeta):

    def __new__(mcs, name, bases, namespace, **kwargs):
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)

        if IconScoreObject in bases or name == "IconSystemScoreBase":
            return cls

        if not isinstance(namespace, dict):
            raise InvalidParamsException('namespace is not dict!')

        elements: Mapping[str, ScoreElementMetadata] = create_score_element_metadatas(cls)
        setattr(cls, CONST_CLASS_ELEMENT_METADATAS, elements)

        api_list = get_score_api(elements.values())
        setattr(cls, CONST_CLASS_API, api_list)

        return cls


class IconScoreBase(IconScoreObject, ContextGetter,
                    metaclass=IconScoreBaseMeta):
    """
    A base class of SCOREs. This class provides facilities and environments to SCORE to run.
    """

    @abstractmethod
    def on_install(self, **kwargs) -> None:
        """
        Invoked when the contract is deployed for the first time,
        and will not be called again on contract update or deletion afterward.
        This is the place where you initialize the state DB.
        """
        super().on_install(**kwargs)

    @abstractmethod
    def on_update(self, **kwargs) -> None:
        """
        Invoked when the contract is deployed for update.
        This is the place where you migrate old states.
        """
        super().on_update(**kwargs)

    @abstractmethod
    def __init__(self, db: 'IconScoreDatabase') -> None:
        """
        A Python init function. Invoked when the contract is loaded at each node.
        Do not put state-changing works in here.
        """
        super().__init__(db)
        self.__db = db
        self.__address = db.address
        self.__owner = IconScoreContextUtil.get_owner(self._context, self.__address)
        self.__icx = None

        elements: ScoreElementMetadataContainer = self.__get_score_element_metadatas()
        if elements.externals == 0:
            raise InvalidExternalException('There is no external method in the SCORE')

        self.__db.set_observer(self.__create_db_observer())

    def fallback(self) -> None:
        """
        fallback function can not be decorated with `@external`.
        (i.e., fallback function is not allowed to be called by external contract or user.)
        This fallback function is executed whenever the contract receives plain icx coins without data.
        If the fallback function is not decorated with `@payable`,
        it is not listed on the SCORE APIs also cannot be called.
        """
        pass

    @classmethod
    def __get_api(cls) -> dict:
        return getattr(cls, CONST_CLASS_API, "")

    def __validate_external_method(self, func_name: str) -> None:
        """Validate the method indicated by func_name is an external method

        :param func_name: name of method
        """

        if not self.__is_external_method(func_name):
            raise MethodNotFoundException(
                f"Method not found: {type(self).__name__}.{func_name}")

    @classmethod
    def __get_score_element_metadatas(cls) -> ScoreElementMetadataContainer:
        return getattr(cls, CONST_CLASS_ELEMENT_METADATAS)

    def __create_db_observer(self) -> 'DatabaseObserver':
        return DatabaseObserver(
            self.__on_db_get, self.__on_db_put, self.__on_db_delete)

    def __call(self,
               func_name: str,
               arg_params: Optional[list] = None,
               kw_params: Optional[dict] = None) -> Any:

        if func_name == STR_FALLBACK:
            if self._context.revision >= Revision.THREE.value:
                if not self.__is_payable_method(func_name):
                    raise MethodNotFoundException(
                        f"Method not found: {type(self).__name__}.{func_name}")
            else:
                self.__check_payable(func_name)

            score_func = getattr(self, func_name)
            ret = score_func()
        else:
            self.__validate_external_method(func_name)
            self.__check_payable(func_name)
            score_func = getattr(self, func_name)
            if arg_params is None:
                arg_params = []
            if kw_params is None:
                kw_params = {}
            ret = score_func(*arg_params, **kw_params)
        return ret

    def __check_payable(self, func_name: str):
        if self.msg.value > 0 and not self.__is_payable_method(func_name):
            raise MethodNotPayableException(
                f"Method not payable: {type(self).__name__}.{func_name}")

    def __is_external_method(self, func_name) -> bool:
        elements = self.__get_score_element_metadatas()
        func: FunctionMetadata = elements.get(func_name)
        return isinstance(func, FunctionMetadata) and func.is_external

    def __is_payable_method(self, func_name) -> bool:
        elements = self.__get_score_element_metadatas()
        func: FunctionMetadata = elements.get(func_name)
        return isinstance(func, FunctionMetadata) and func.is_payable

    def __is_func_readonly(self, func_name: str) -> bool:
        elements = self.__get_score_element_metadatas()
        func: FunctionMetadata = elements.get(func_name)
        return isinstance(func, FunctionMetadata) and func.is_readonly

    # noinspection PyUnusedLocal
    @staticmethod
    def __on_db_get(context: 'IconScoreContext',
                    key: bytes,
                    value: bytes):
        """Invoked when `get` is called in `ContextDatabase`.

        # All steps are managed in the score
        # Don't move to another codes

        :param context: SCORE context
        :param key: key
        :param value: value
        """

        if context and context.step_counter and \
                context.type != IconScoreContextType.DIRECT:
            length = 1
            if value:
                length = len(value)
            context.step_counter.apply_step(StepType.GET, length)

    # noinspection PyUnusedLocal
    @staticmethod
    def __on_db_put(context: 'IconScoreContext',
                    key: bytes,
                    old_value: bytes,
                    new_value: bytes):
        """Invoked when `put` is called in `ContextDatabase`.

        # All steps are managed in the score
        # Don't move to another codes

        :param context: SCORE context
        :param key: key
        :param old_value: old value
        :param new_value: new value
        """

        if context and context.step_counter and not context.readonly:
            if old_value:
                # modifying a value
                context.step_counter.apply_step(
                    StepType.REPLACE, len(new_value))
            else:
                # newly storing a value
                context.step_counter.apply_step(
                    StepType.SET, len(new_value))

    # noinspection PyUnusedLocal
    @staticmethod
    def __on_db_delete(context: 'IconScoreContext',
                       key: bytes,
                       old_value: bytes):
        """Invoked when `delete` is called in `ContextDatabase`.

        # All steps are managed in the score
        # Don't move to another codes
        :param context: SCORE context
        :param key: key
        :param old_value: old value
        """

        if context and context.step_counter and not context.readonly:
            context.step_counter.apply_step(
                StepType.DELETE, len(old_value))

    @property
    def msg(self) -> 'Message':
        """
        Holds information of calling the SCORE

        -  msg.sender : Address of the account who called this function. If
           other contact called this function, msg.sender points to the caller
           contract’s address.

        -  msg.value : Amount of icx that the sender attempts to transfer to the
           current SCORE.
        """
        return self._context.msg

    @property
    def address(self) -> 'Address':
        """
        The current SCORE address

        :return: :class:`.Address` current address
        """
        return self.__address

    @property
    def tx(self) -> 'Transaction':
        """
        Holds information of the transaction

        :return: :class:`.Transaction` transaction
        """
        return self._context.tx

    @property
    def block(self) -> 'Block':
        """
        Deprecated property

        Use block_height and now() instead.
        """
        warnings.warn("Use block_height and now() instead", DeprecationWarning, stacklevel=2)
        return Block(self._context.block.height, self._context.block.timestamp)

    @property
    def db(self) -> 'IconScoreDatabase':
        """
        An instance used to access state DB

        :return: :class:`.IconScoreDatabase` db
        """
        return self.__db

    @property
    def owner(self) -> 'Address':
        """
        Address of the account who deployed the contract

        :return: :class:`.Address` owner address
        """
        return self.__owner

    @property
    def icx(self) -> 'Icx':
        """
        An object used to transfer icx coin

        -  icx.transfer(addr_to(address), amount(integer)) -> bool Transfers
           designated amount of icx coin to ``addr_to``. If exception occurs
           during execution, the exception will be escalated. Returns True if
           coin transfer succeeds.

        -  icx.send(addr_to(address), amount(integer)) -> bool Sends designated
           amount of icx coin to ``addr_to``. Basic behavior is same as
           transfer, the difference is that exception is caught inside the
           function. Returns True when coin transfer succeeded, False when
           failed.

        :return: :class:`.Icx` instance of icx
        """
        if self.__icx is None:
            self.__icx = Icx(self._context, self.__address)
        else:
            # Should update a new context in icx for every tx
            self.__icx._context = self._context

        return self.__icx

    @property
    def block_height(self) -> int:
        """
        Current block height

        :return: current block height
        """
        return self._context.block.height

    def now(self) -> int:
        """
        Timestamp of current block in microseconds

        :return: timestamp in microseconds
        """
        return self._context.block.timestamp

    def call(self, addr_to: 'Address', func_name: str, kw_dict: dict, amount: int = 0):
        """
        Calls an external function provided by another SCORE.
        `func_name` can be `None` if fallback calls

        :param addr_to: :class:`.Address` the address of another SCORE
        :param func_name: function name of another SCORE
        :param kw_dict: arguments of the external function
        :param amount: amount of ICX to transfer in loop
        :return: returning value of the external function
        """
        return InternalCall.other_external_call(self._context, self.address, addr_to, amount, func_name, (), kw_dict)

    @staticmethod
    def revert(message: Optional[str] = None, code: int = 0):
        """
        Deprecated method

        Use global function `revert()` instead.
        """
        warnings.warn("Use global function revert() instead.", DeprecationWarning, stacklevel=2)
        revert(message, code)

    def is_score_active(self, score_address: 'Address') -> bool:
        warnings.warn("Forbidden function", DeprecationWarning, stacklevel=2)
        if self._context.revision <= Revision.THREE.value:
            return IconScoreContextUtil.is_score_active(self._context, score_address)
        else:
            raise AccessDeniedException('No permission')

    def get_owner(self, score_address: Optional['Address']) -> Optional['Address']:
        warnings.warn("Forbidden function", DeprecationWarning, stacklevel=2)
        if self._context.revision <= Revision.THREE.value:
            return IconScoreContextUtil.get_owner(self._context, score_address)
        else:
            raise AccessDeniedException('No permission')

    @staticmethod
    def create_interface_score(addr_to: 'Address',
                               interface_cls: Callable[['Address'], T]) -> T:
        """
        Creates an object, through which you have an access to the designated SCORE’s external functions.

        :param addr_to: SCORE address
        :param interface_cls: interface class
        :return: An instance of given class
        """

        if interface_cls is InterfaceScore:
            raise InvalidInstanceException(FORMAT_IS_NOT_DERIVED_OF_OBJECT.format(InterfaceScore.__name__))
        return interface_cls(addr_to)

    def deploy(self, tx_hash: bytes):
        warnings.warn("Forbidden function", DeprecationWarning, stacklevel=2)
        if self._context.revision <= Revision.THREE.value and \
                self.address == GOVERNANCE_SCORE_ADDRESS:
            # switch
            score_addr: 'Address' = self.get_score_address_by_tx_hash(tx_hash)
            owner: 'Address' = self.get_owner(score_addr)
            tmp_sender: 'Address' = self._context.msg.sender

            self._context.msg.sender = owner
            try:
                IconScoreContextUtil.deploy(self._context, tx_hash)
            finally:
                self._context.msg.sender = tmp_sender
        else:
            raise AccessDeniedException('No permission')

    def get_tx_hashes_by_score_address(self,
                                       score_address: 'Address') -> Tuple[Optional[bytes], Optional[bytes]]:
        warnings.warn("Forbidden function", DeprecationWarning, stacklevel=2)
        if self._context.revision <= Revision.THREE.value:
            return IconScoreContextUtil.get_tx_hashes_by_score_address(self._context, score_address)
        else:
            raise AccessDeniedException('No permission')

    def get_score_address_by_tx_hash(self,
                                     tx_hash: bytes) -> Optional['Address']:
        warnings.warn("Forbidden function", DeprecationWarning, stacklevel=2)
        if self._context.revision <= Revision.THREE.value:
            return IconScoreContextUtil.get_score_address_by_tx_hash(self._context, tx_hash)
        else:
            raise AccessDeniedException('No permission')

    def get_fee_sharing_proportion(self):
        return self._context.fee_sharing_proportion

    def set_fee_sharing_proportion(self, proportion: int):
        if self._context.tx.to == self.address:
            if self._context.type == IconScoreContextType.QUERY:
                raise InvalidRequestException("Cannot set fee sharing proportion in read-only context")
            if proportion < 0 or proportion > 100:
                raise InvalidRequestException("Invalid proportion: should be between 0 and 100")

            self._context.fee_sharing_proportion = proportion



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import IntEnum, unique
from typing import Optional


@unique
class ExceptionCode(IntEnum):
    OK = 0
    SYSTEM_ERROR = 1
    CONTRACT_NOT_FOUND = 2
    METHOD_NOT_FOUND = 3
    METHOD_NOT_PAYABLE = 4
    ILLEGAL_FORMAT = 5
    INVALID_PARAMETER = 6
    INVALID_INSTANCE = 7
    INVALID_CONTAINER_ACCESS = 8
    ACCESS_DENIED = 9
    OUT_OF_STEP = 10
    OUT_OF_BALANCE = 11
    TIMEOUT_ERROR = 12
    STACK_OVERFLOW = 13
    INVALID_PACKAGE = 14
    SERVICE_NOT_READY = 15
    INTERNAL_SERVICE_ERROR = 16
    INVALID_BALANCE = 17

    # Caused by revert call or user-defined exception.
    SCORE_ERROR = 32
    END = 99

    def __str__(self) -> str:
        return str(self.name).capitalize().replace('_', ' ')


class FatalException(BaseException):
    # if this exception is raised on invoke or commit, icon service will be closed
    pass


class InvalidBaseTransactionException(BaseException):
    pass


class IconServiceBaseException(BaseException):
    """All custom exceptions used in IconService should inherit from this
    """

    def __init__(self, message: Optional[str], code: ExceptionCode = ExceptionCode.OK):
        if message is None:
            message = str(code)
        self.__message = message
        self.__code = code

    @property
    def message(self):
        return self.__message

    @property
    def code(self):
        return self.__code

    def __str__(self):
        return f'{self.message} ({self.code})'


class ScoreNotFoundException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.CONTRACT_NOT_FOUND)


class MethodNotFoundException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.METHOD_NOT_FOUND)


class MethodNotPayableException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.METHOD_NOT_PAYABLE)


class InvalidParamsException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.INVALID_PARAMETER)


class AccessDeniedException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.ACCESS_DENIED)


class DatabaseException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.ACCESS_DENIED)


class InvalidInstanceException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.INVALID_INSTANCE)


class InvalidContainerAccessException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.INVALID_CONTAINER_ACCESS)


class IllegalFormatException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.ILLEGAL_FORMAT)


class InvalidRequestException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.ILLEGAL_FORMAT)


class InvalidExternalException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.ILLEGAL_FORMAT)


class InvalidPayableException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.ILLEGAL_FORMAT)


class InvalidEventLogException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.ILLEGAL_FORMAT)


class InvalidInterfaceException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.ILLEGAL_FORMAT)


class OutOfBalanceException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.OUT_OF_BALANCE)


class TimeoutException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.TIMEOUT_ERROR)


class StackOverflowException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.STACK_OVERFLOW)


class InvalidPackageException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.INVALID_PACKAGE)


class ServiceNotReadyException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.SERVICE_NOT_READY)


class InternalServiceErrorException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.INTERNAL_SERVICE_ERROR)


class InvalidBalanceException(IconServiceBaseException):
    def __init__(self, message: Optional[str]):
        super().__init__(message, ExceptionCode.INVALID_BALANCE)


class IconScoreException(IconServiceBaseException):
    # All the user-defined exceptions should inherit from this exception including revert call
    def __init__(self, message: Optional[str], index: int = 0):
        if not isinstance(index, int):
            raise InvalidParamsException('Invalid index type: not an integer')
        code = ExceptionCode.SCORE_ERROR + index
        if code < ExceptionCode.SCORE_ERROR:
            code = ExceptionCode.SCORE_ERROR
        elif code > ExceptionCode.END:
            code = ExceptionCode.END
        super().__init__(message, code)



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, List, Optional, Any

from .icon_score_step import StepType
from ..base.address import Address, ICON_ADDRESS_BYTES_SIZE, ICON_ADDRESS_BODY_SIZE
from ..base.exception import InvalidEventLogException
from ..icon_constant import DATA_BYTE_ORDER, Revision
from ..utils import int_to_bytes, byte_length_of_int

if TYPE_CHECKING:
    from .icon_score_constant import BaseType
    from .icon_score_context import IconScoreContext


class EventLog(object):
    """ A DataClass of a event log.
    """

    def __init__(
            self,
            score_address: 'Address',
            indexed: List['BaseType'],
            data: List['BaseType']) -> None:
        """
        Constructor

        :param score_address: an address of SCORE in which the event is invoked
        :param indexed: a list of indexed arguments including a event signature
        :param data: a list of normal arguments
        """

        assert score_address.is_contract
        assert isinstance(indexed, list)
        assert isinstance(data, list)

        self.score_address: 'Address' = score_address
        self.indexed: 'List[BaseType]' = indexed
        self.data: 'List[BaseType]' = data

    def __str__(self) -> str:
        return '\n'.join([f'{k}: {v}' for k, v in self.__dict__.items()])

    def to_dict(self, casing: Optional[callable] = None) -> dict:
        """
        Returns properties as `dict`
        :return: a dict
        """
        new_dict = {}
        for key, value in self.__dict__.items():
            if value is None:
                # Excludes properties which have `None` value
                continue

            new_dict[casing(key) if casing else key] = value

        return new_dict


class EventLogEmitter:
    @classmethod
    def emit_event_log(cls,
                       context: 'IconScoreContext',
                       score_address: 'Address',
                       event_signature: str,
                       arguments: List[Any],
                       indexed_args_count: int,
                       fee_charge: bool = False):
        """
        Puts a eventlog to the running context

        :param context: running context.
        :param score_address: score address which event is occurred at.
        :param event_signature: signature of the eventlog
        :param arguments: arguments of eventlog call
        :param indexed_args_count: count of the indexed arguments
        :param fee_charge: used for deciding whether fee will be charged for the event logs

        :return:
        """

        if context.readonly:
            raise InvalidEventLogException(
                'The event log can not be recorded on readonly context')

        if indexed_args_count > len(arguments):
            raise InvalidEventLogException(
                f'declared indexed_args_count is {indexed_args_count}, '
                f'but argument count is {len(arguments)}')

        event_size = EventLogEmitter.__get_byte_length(context, event_signature)
        indexed: List['BaseType'] = [event_signature]
        data: List['BaseType'] = []
        for i, argument in enumerate(arguments):
            event_size += EventLogEmitter.__get_byte_length(context, argument)

            # Separates indexed type and base type with keeping order.
            if i < indexed_args_count:
                indexed.append(argument)
            else:
                data.append(argument)

        # Counting steps only if fee_charge is True
        if fee_charge:
            context.step_counter.apply_step(StepType.EVENT_LOG, event_size)

        event = EventLog(score_address, indexed, data)
        context.event_logs.append(event)

    @classmethod
    def __get_byte_length(cls,
                          context: 'IconScoreContext',
                          data: 'BaseType') -> int:
        if data is None:
            return 0
        elif isinstance(data, int):
            return byte_length_of_int(data)
        elif isinstance(data, Address):
            if context.revision < Revision.THREE.value:
                return ICON_ADDRESS_BODY_SIZE
            else:
                return ICON_ADDRESS_BYTES_SIZE

        return len(cls.__get_bytes_from_base_type(data))

    @classmethod
    def __get_bytes_from_base_type(cls,
                                   data: 'BaseType') -> bytes:
        if isinstance(data, str):
            return data.encode('utf-8')
        elif isinstance(data, Address):
            return data.prefix.value.to_bytes(1, DATA_BYTE_ORDER) + data.body
        elif isinstance(data, bytes):
            return data
        elif isinstance(data, int):
            return int_to_bytes(data)
        else:
            raise InvalidEventLogException(f'Invalid data type: {type(data)}, data: {data}')

    @classmethod
    def get_ordered_bytes(cls,
                          index: int,
                          data: 'BaseType') -> bytes:
        bloom_data = index.to_bytes(1, DATA_BYTE_ORDER)
        if data is not None:
            bloom_data += cls.__get_bytes_from_base_type(data)
        return bloom_data
